package ru.dbotthepony.mc.otm.container

import net.minecraft.world.item.ItemStack
import net.neoforged.neoforge.capabilities.Capabilities
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.fluid.stream
import ru.dbotthepony.mc.otm.core.isNotEmpty
import ru.dbotthepony.mc.otm.core.math.Decimal

interface HandlerFilter {
	fun canInsert(slot: Int, stack: ItemStack): Boolean {
		return true
	}

	fun canExtract(slot: Int, amount: Int, stack: ItemStack): Boolean {
		return true
	}

	fun preInsert(slot: Int, stack: ItemStack, simulate: Boolean) {}
	fun preExtract(slot: Int, amount: Int, simulate: Boolean) {}

	fun modifyInsertCount(slot: Int, stack: ItemStack, existing: ItemStack, simulate: Boolean): Int {
		return stack.count
	}

	fun modifyExtractCount(slot: Int, amount: Int, simulate: Boolean): Int {
		return amount
	}

	fun and(other: HandlerFilter): HandlerFilter {
		return object : HandlerFilter {
			override fun canInsert(slot: Int, stack: ItemStack): Boolean {
				return this@HandlerFilter.canInsert(slot, stack) && other.canInsert(slot, stack)
			}

			override fun canExtract(slot: Int, amount: Int, stack: ItemStack): Boolean {
				return this@HandlerFilter.canExtract(slot, amount, stack) && other.canExtract(slot, amount, stack)
			}

			override fun preInsert(slot: Int, stack: ItemStack, simulate: Boolean) {
				this@HandlerFilter.preInsert(slot, stack, simulate)
				other.preInsert(slot, stack, simulate)
			}

			override fun preExtract(slot: Int, amount: Int, simulate: Boolean) {
				this@HandlerFilter.preExtract(slot, amount, simulate)
				other.preExtract(slot, amount, simulate)
			}
		}
	}

	object FluidContainers : HandlerFilter {
		override fun canInsert(slot: Int, stack: ItemStack): Boolean {
			return stack.getCapability(Capabilities.FluidHandler.ITEM)?.let { it.tanks > 0 } ?: false
		}
	}

	object DrainableFluidContainers : HandlerFilter {
		override fun canInsert(slot: Int, stack: ItemStack): Boolean {
			return stack.getCapability(Capabilities.FluidHandler.ITEM)?.let { it.stream().anyMatch { it.isNotEmpty } } ?: false
		}

		override fun canExtract(slot: Int, amount: Int, stack: ItemStack): Boolean {
			return !canInsert(slot, stack)
		}
	}

	object OnlyIn : HandlerFilter {
		override fun canInsert(slot: Int, stack: ItemStack): Boolean {
			return true
		}

		override fun canExtract(slot: Int, amount: Int, stack: ItemStack): Boolean {
			return false
		}
	}

	object OnlyOut : HandlerFilter {
		override fun canInsert(slot: Int, stack: ItemStack): Boolean {
			return false
		}

		override fun canExtract(slot: Int, amount: Int, stack: ItemStack): Boolean {
			return true
		}
	}

	object Both : HandlerFilter

	object Dischargeable : HandlerFilter {
		override fun canInsert(slot: Int, stack: ItemStack): Boolean {
			return stack.getCapability(Capabilities.EnergyStorage.ITEM)?.let { it.canExtract() && it.extractEnergy(Int.MAX_VALUE, true) > 0 } ?: false
		}

		override fun canExtract(slot: Int, amount: Int, stack: ItemStack): Boolean {
			return stack.getCapability(Capabilities.EnergyStorage.ITEM)?.let { !it.canExtract() || it.extractEnergy(Int.MAX_VALUE, true) <= 0 } ?: false
		}
	}

	object Chargeable : HandlerFilter {
		override fun canInsert(slot: Int, stack: ItemStack): Boolean {
			return stack.getCapability(Capabilities.EnergyStorage.ITEM)?.let { it.canReceive() && it.receiveEnergy(Int.MAX_VALUE, true) > 0 } ?: false
		}

		override fun canExtract(slot: Int, amount: Int, stack: ItemStack): Boolean {
			return stack.getCapability(Capabilities.EnergyStorage.ITEM)?.let { !it.canReceive() || it.receiveEnergy(Int.MAX_VALUE, true) <= 0 } ?: false
		}
	}

	object ChemicalFuel : HandlerFilter {
		override fun canExtract(slot: Int, amount: Int, stack: ItemStack): Boolean {
			return stack.getBurnTime(null) <= 0
		}

		override fun canInsert(slot: Int, stack: ItemStack): Boolean {
			return stack.getBurnTime(null) > 0
		}
	}

	object IsPattern : HandlerFilter {
		override fun canInsert(slot: Int, stack: ItemStack): Boolean {
			return stack.getCapability(MatteryCapability.PATTERN_ITEM) != null
		}
	}

	object MatterProviders : HandlerFilter {
		override fun canInsert(slot: Int, stack: ItemStack): Boolean {
			return stack.getCapability(MatteryCapability.MATTER_ITEM)
				?.let { it.matterFlow.output && it.extractMatterChecked(Decimal.POSITIVE_INFINITY, true) > Decimal.ZERO }
				?: false
		}

		override fun canExtract(slot: Int, amount: Int, stack: ItemStack): Boolean {
			return stack.getCapability(MatteryCapability.MATTER_ITEM)
				?.let { !it.matterFlow.output || it.extractMatterChecked(Decimal.POSITIVE_INFINITY, true) <= Decimal.ZERO }
				?: false
		}
	}

	object MatterConsumers : HandlerFilter {
		override fun canInsert(slot: Int, stack: ItemStack): Boolean {
			return stack.getCapability(MatteryCapability.MATTER_ITEM)
				?.let { it.matterFlow.input && it.receiveMatterChecked(Decimal.POSITIVE_INFINITY, true) > Decimal.ZERO }
				?: false
		}

		override fun canExtract(slot: Int, amount: Int, stack: ItemStack): Boolean {
			return stack.getCapability(MatteryCapability.MATTER_ITEM)
				?.let { !it.matterFlow.input || it.receiveMatterChecked(Decimal.POSITIVE_INFINITY, true) <= Decimal.ZERO }
				?: false
		}
	}
}
